import pyawr.mwoffice as mwo
from pyawr_utils import common_utils
from tkinter import messagebox
from tkinter import filedialog
import warnings
import json
import numpy as np
#
_Elements = common_utils._Elements
_Equations = common_utils._Equations
#
#*****************************************************************************
#
class _Goal():
    def __init__(self, awrde, goal_idx: int, goal_source: str):#------------------------------
        '''
        goal_source: Optimize|Yield
        '''
        self.awrde = awrde
        self._initialize_goal(goal_idx, goal_source)
        #
    def _initialize_goal(self, goal_idx, goal_source):#----------------------------------------
        try:
            if goal_source == 'Optimize':
                self._goal = self.awrde.Project.OptGoals(goal_idx+1)
            elif goal_source == 'Yield':
                self._goal = self.awrde.Project.YieldGoals(goal_idx+1)
            #end if
        except:
            raise RuntimeError('Goal does not exist: ' + goal_name)
        #end try
        self._goal_name = self._goal.Name
        #
    @property
    def goal_name(self) -> str:#-----------------------------------------------------
        return self._goal_name
        #
    @property
    def goal_enabled(self) -> bool:#--------------------------------------------------
        return self._goal.Enable
        #
    @goal_enabled.setter
    def goal_enabled(self, enabled_state:bool):#--------------------------------------------------
        self._goal.Enable = enabled_state
        #
    @property
    def goal_xrange(self) -> tuple:#-----------------------------------------------
        xStart = self._goal.xStart
        xStop = self._goal.xStop
        return (xStart, xStop)
        #
    @goal_xrange.setter
    def goal_xrange(self, xrange_tuple: tuple):#-----------------------------------------------
        if not isinstance(xrange_tuple, tuple):
            raise TypeError('xrange must be tuple: (xStart, xStop)')
        #end if
        try:
            xStart = xrange_tuple[0]
            xStop = xrange_tuple[1]
            self._goal.xStart = xStart
            self._goal.xStop = xStop
        except:
            warnings.warn('goal_xrange could not be set')
        #end try
        #
    @property
    def goal_yrange(self) -> tuple:#-----------------------------------------------
        yStart = self._goal.yStart
        yStop = self._goal.yStop
        return (yStart, yStop)
        #
    @goal_yrange.setter
    def goal_yrange(self, yrange_tuple: tuple):#-----------------------------------------------
        if not isinstance(yrange_tuple, tuple):
            raise TypeError('yrange must be tuple: (yStart, yStop)')
        #end if
        try:
            yStart = yrange_tuple[0]
            yStop = yrange_tuple[1]
            self._goal.yStart = yStart
            self._goal.yStop = yStop
        except:
            warnings.warn('goal_xrange could not be set')
        #end try
        #
    @property
    def goal_type(self) -> str:#--------------------------------------------------------
        '''
        returns the goal type as a string: =, >, or <
        
        The API only allows reading the Type. The API does not allow setting the type.
        The user would need to add a new goal with the desired type and delete the old goal
        
        '''
        goal_type_val = self._goal.Type
        if goal_type_val == 0:
            goal_type_str = '='
        elif goal_type_val == 1:
            goal_type_str = '<'
        elif goal_type_val == 2:
            goal_type_str = '>'
        else:
            warnings.warn('invalid goal type')
            goal_type_str = 'NA'
        #end if
        return goal_type_str
        #
    @property
    def goal_weight(self) -> float:#---------------------------------------------------
        return self._goal.Weight
        #
    @goal_weight.setter
    def goal_weight(self, goal_weight: float):#----------------------------------------
        if (not isinstance(goal_weight, float)) and (not isinstance(goal_weight, int)):
            raise TypeError('goal_weight must be either float or int type')
        #end if
        self._goal.Weight = goal_weight
        #
    @property
    def goal_l_factor(self) -> float:#----------------------------------------------
        return self._goal.LVal
        #
    @goal_l_factor.setter
    def goal_l_factor(self, goal_l_factor: float):#-------------------------------------
        if (not isinstance(goal_l_factor, float)) and (not isinstance(goal_l_factor, int)):
            raise TypeError('goal_l_factor must be either float or int type')
        #end if
        self._goal.LVal = goal_l_factor
        #
    @property
    def goal_cost(self) -> float:#---------------------------------------------------
        return self._goal.Cost
    
#
#*****************************************************************************
#
class _DataFiles():
    '''
    Methods for collection of Data Files in the AWRDE project
    
    Parameters
    ----------
    awrde: The AWRDE object returned from awrde_utils.establish_link()
    
    '''
    def __init__(self, awrde):#----------------------------------------
        self.awrde = awrde
        #
    def _build_data_files_dict(self):#--------------------------------
        '''Create list of all graph names in project
    
        Parameters
        ----------
        None
    
        Returns
        -------
        _data_files_dict: dict
                         keys are the data file names
                         values are the data file objects
        '''        
        self._data_files_dict = {}
        for df in self.awrde.Project.DataFiles:
            data_file = _DataFile(self.awrde, df.Name)
            self._data_files_dict[df.Name] = data_file
        #end for
        #
    @property
    def data_files_dict(self) -> dict:#-----------------------------------------------
        '''return data file dictionary'''
        self._build_data_files_dict()
        return self._data_files_dict
        #
    @property
    def data_file_names_list(self) -> list[str]:#-------------------------------------------
        '''returns list of data file names'''
        data_file_name_list = list()
        for df in self.awrde.Project.DataFiles:
            data_file_name_list.append(df.Name)
        #end for
        return data_file_name_list
        #
    def add_dc_iv_file(self, data_file_name: str):#-----------------------------------
        '''
        Add a DC-IV file
        
        Parameters
        ----------
        data_file_name: string
        '''
        self._add_data_file(data_file_name, 'DC-IV')
        #
    def add_dscr_file(self, data_file_name: str):#-----------------------------------
        '''
        Add a DSCR file
        
        Parameters
        ----------
        data_file_name: string
        '''
        self._add_data_file(data_file_name, 'DSCR')
        #
    def add_gmdif_file(self, data_file_name: str):#-----------------------------------
        '''
        Add a Generalized MDIF file
        
        Parameters
        ----------
        data_file_name: string
        '''
        self._add_data_file(data_file_name, 'GMDIF')
        #
    def add_gmdif_nport_file(self, data_file_name: str):#-----------------------------------
        '''
        Add a Generalized MDIF N-Port file
        
        Parameters
        ----------
        data_file_name: string
        '''
        self._add_data_file(data_file_name, 'GMDIF N-Port')
        #
    def add_mdif_file(self, data_file_name: str):#-----------------------------------
        '''
        Add a MDIF file
        
        Parameters
        ----------
        data_file_name: string
        '''
        self._add_data_file(data_file_name, 'MDIF')
        #
    def add_raw_port_parameters_file(self, data_file_name: str):#-----------------------------------
        '''
        Add a Raw Port Parameters file
        
        Parameters
        ----------
        data_file_name: string
        '''
        self._add_data_file(data_file_name, 'Raw Port Parameters')
        #
    def add_text_file(self, data_file_name: str):#-----------------------------------
        '''
        Add a Text file
        
        Parameters
        ----------
        data_file_name: string
        '''
        self._add_data_file(data_file_name, 'Text')
        #
    def add_touchstone_file(self, data_file_name: str):#-----------------------------------
        '''
        Add a Touchstone file
        
        Parameters
        ----------
        data_file_name: string
        '''
        self._add_data_file(data_file_name, 'Touchstone')
        #
    def remove_data_file(self, data_file_name: str) -> bool:
        '''
        Delete a data file from the AWRDE project
        
        Parameters
        ----------
        data_file_name: string,
                 name of the Data File to be deleted
        
        Returns
        -------
        data_file_removed: bool
                    True if data file successfully deleted,
                    False if data file could not be deleted
                    
        '''        
        if not isinstance(data_file_name, str):
            raise TypeError('data_file_name must be a string type')
        #end if
        data_file_removed = False               
        try:
            for df_idx in range(self.awrde.Project.DataFiles.Count):
                if self.awrde.Project.DataFiles[df_idx].Name == data_file_name:
                    self.awrde.Project.DataFiles.Remove(df_idx+1)
                    data_file_removed = True
                    break
                #end if
            #end for
        except:
            warnings.warn('remove_data_file: Data File ' + data_file_name + ' did not get removed')
        #end for
        return data_file_removed
        #
    def rename_data_file(self, data_file_name: str, new_data_file_name: str):#--------------------
        '''
        rename Data File
    
        Parameters
        ----------
        data_file_name: string
                         existing graph name
        new_data_file_name: string
                         new graph name
        '''        
        if not isinstance(data_file_name, str):
            raise TypeError('data_file_name must be a string')
        #end if
        if not isinstance(new_data_file_name, str):
            raise TypeError('new_data_file_name must be a string')
        #end if
        data_file_name_list = self.data_file_names_list
        found_it = False
        for data_file_name_from_list in data_file_name_list:
            if data_file_name_from_list == data_file_name:
                found_it = True
                if new_data_file_name in data_file_name_list:
                    YesNo = messagebox.askyesno('Rename Data File','Data File '\
                        + new_data_file_name + ' exists. Remove existing Data File ?')
                    if YesNo:
                        self.remove_data_file(new_data_file_name)
                    else:
                        new_data_file_name += ' 1'
                    #end if
                #end if
                try:
                    df = self.awrde.Project.DataFiles(data_file_name)
                    df.Name = new_data_file_name
                except:
                    raise RuntimeError('problem with renaming data file')
                #end try
            #end if
        #end for
        if not found_it:
            warnings.warn('data file rename failed: ' + data_file_name)
        #end if
        #
    def copy_data_file(self, data_file_name: str, new_data_file_name: str):#-----------------------
        '''
        copy Data File
    
        Parameters
        ----------
        data_file_name: string
                         existing data file name
        new_data_file_name: string
                         new data file name
        '''
        if not isinstance(data_file_name, str):
            raise TypeError('data_file_name must be a string')
        #end if
        if not isinstance(new_data_file_name, str):
            raise TypeError('new_data)file_name must be a string')
        #end if        
        for df in self.awrde.Project.DataFiles:
            if df.Name == new_data_file_name:
                new_data_file_name = new_data_file_name + ' 1'
            #end if
        #end for
        #
        try:
            for df_idx in range(1, self.awrde.Project.DataFiles.Count+1):
                df = self.awrde.Project.DataFiles(df_idx)
                if df.Name == data_file_name:
                    self.awrde.Project.DataFiles.Copy(df_idx, new_data_file_name)
                #end if
            #end for
        except:
            raise RuntimeError('problem with copying data file')
        #end try
        #
    def import_data_file(self, data_file_name: str, file_path: str, data_file_type: str):#------
        '''
        import data file
        
        Parameters
        ----------
        data_file_name: string
               name of the data file as it will appear in the project
               
        file_path: string
               full directory and file name of the file to be imported
               
        data_file_type: string
               valid values:
                    DC-IV File
                    DSCR File
                    Generalized MDIF Data File
                    Generalized MDIF N-Port File
                    MDIF File
                    Raw Port Parameters File
                    Text File
                    Touchstone File
        '''
        if not isinstance(data_file_name, str):
            raise TypeError('data_file_name must be string type')
        #end if
        if not isinstance(file_path, str):
            raise TypeError('file_path must be string type')
        #end if
        if not isinstance(data_file_type, str):
            raise TypeError('data_file_type must be string type')
        #end if
        #
        if data_file_type == 'DC-IV File':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_IV
        elif data_file_type == 'DSCR File':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_DSCR
        elif data_file_type == 'Generalized MDIF Data File':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_GMDIF
        elif data_file_type == 'Generalized MDIF N-Port File':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_GMDIFD
        elif data_file_type == 'MDIF File':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_MDIF
        elif data_file_type == 'Raw Port Parameters File':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_RAW
        elif data_file_type == 'Text File':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_TXT
        elif data_file_type == 'Touchstone File':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_SNP
        else:
            raise RuntimeError('data_file_type not recognized: ' + data_file_type)
        #end if
        if self._does_data_file_exist(data_file_name):
            YesNo = messagebox.askyesno('Import Data File','Data File ' + data_file_name + ' exists. Delete existing Data File?')
            if YesNo:
                self.remove_data_file(data_file_name)
            #end if
        #end if
        try:
            self.awrde.Project.DataFiles.Import(data_file_name, file_path, data_file_type_enum)
        except:
            raise RuntimeError('Problem with importing data file')
        #end try
        #
    def _does_data_file_exist(self, data_file_name: str) -> bool:#------------------------
        data_file_exists = False
        for df in self.awrde.Project.DataFiles:
            if df.Name == data_file_name:
                data_file_exists = True
            #end if
        #end for
        return data_file_exists
        #
    def _add_data_file(self, data_file_name, data_file_type):#----------------------------
        if data_file_type == 'DC-IV':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_IV
        elif data_file_type == 'DSCR':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_DSCR
        elif data_file_type == 'GMDIF':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_GMDIF
        elif data_file_type == 'GMDIF N-Port':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_GMDIFD
        elif data_file_type == 'MDIF':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_MDIF
        elif data_file_type == 'Raw Port Parameters':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_RAW
        elif data_file_type == 'Text':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_TXT
        elif data_file_type == 'Touchstone':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_SNP
        #end if
        if self._does_data_file_exist(data_file_name):
            YesNo = messagebox.askyesno('Add Data File','Data File ' + data_file_name + ' exists. Delete existing Data File?')
            if YesNo:
                self.remove_data_file(data_file_name)
            else:
                data_file_name += ' 1'
            #end if
        #end if
        try:
            self.awrde.Project.DataFiles.AddNew(data_file_name, data_file_type_enum)
        except:
            warnings.warn('Data File not added. Possible name conflict')
        #end try
#
#*****************************************************************************
#
class _DataFile():
    def __init__(self, awrde, data_file_name: str):#-----------------------------------
        self.awrde = awrde
        self._initialize_data_file(data_file_name)
        #
    def _initialize_data_file(self, data_file_name: str):#-------------------------------------
        try:
            self._df = self.awrde.Project.DataFiles(data_file_name)
        except:
            raise RuntimeError('Data File does not exist: ' + data_file_name)
        #end try
        #
    @property
    def data_file_name(self) -> str:#------------------------------------------------------
        '''Returns name of the Data File'''
        return self._df.Name
        #
    @property
    def data_file_type(self) -> str:#--------------------------------------------------
        '''
        Reports data file type 
        
        Returns
        -------
        data_file_type_str: string
              DC-IV File, DSCR File, Generalized MDIF Data File, Generalized MDIF N-Port File,
              MDIF File, Raw Port Parameters File, Text File, Touchstone File
        
        ''' 
        data_file_type_enum = self._df.Type
        if mwo.mwDataFileType(data_file_type_enum) == mwo.mwDataFileType.mwDFT_IV:
            data_file_type_str = 'DC-IV File'
        elif mwo.mwDataFileType(data_file_type_enum) == mwo.mwDataFileType.mwDFT_DSCR:
            data_file_type_str = 'DSCR File'
        elif mwo.mwDataFileType(data_file_type_enum) == mwo.mwDataFileType.mwDFT_GMDIF:
            data_file_type_str = 'Generalized MDIF Data File'
        elif mwo.mwDataFileType(data_file_type_enum) == mwo.mwDataFileType.mwDFT_GMDIFD:
            data_file_type_str = 'Generalized MDIF N-Port File'
        elif mwo.mwDataFileType(data_file_type_enum) == mwo.mwDataFileType.mwDFT_MDIF:
            data_file_type_str = 'MDIF File'
        elif mwo.mwDataFileType(data_file_type_enum) == mwo.mwDataFileType.mwDFT_RAW:
            data_file_type_str = 'Raw Port Parameters File'
        elif mwo.mwDataFileType(data_file_type_enum) == mwo.mwDataFileType.mwDFT_TXT:
            data_file_type_str = 'Text File'
        elif mwo.mwDataFileType(data_file_type_enum) == mwo.mwDataFileType.mwDFT_SNP:
            data_file_type_str = 'Touchstone File'
        #end if
        return data_file_type_str
        #
    def export_data_file(self, file_path: str) -> bool:#--------------------------------------
        '''
        exports the data file to the directory and name specified by file_path
        
        Parameters
        ----------
        file_path: string
              full director and file name
        
        Returns
        -------
        export_successful: bool
        '''
        if not isinstance(file_path,str):
            raise TypeError('file_path must be string type')
        #end if
        #
        export_successful = True
        try:
            self._df.Export(file_path)
        except:
            export_successful = False
        #end try
        return export_successful
        
#
#*****************************************************************************
#
class _Optimization():
    '''
    methods for using the Optimizer
    Includes: setting/modifying goals
              modifying optimization variables
              setting optimizer
              running optimizer
    
    '''
    def __init__(self, awrde):#--------------------------------------------------------
        self.awrde = awrde
        #
    @property
    def optimization_variables_dict(self) -> dict:#-----------------------------------
        '''
        Returns dictionary of variables that are setup for optimization
        
        Returns
        -------
        optimization_variables_dict: dictionary
            Keys: Index
            Values: Dictionary of fields:
                        Document Name
                        Document Type: Schematic, System Diagram, EM Structure, Global Definition
                        Element Type: Element or Equation
                        Element Name
                        Variable Name
                        Nominal Value
                        Constrained: True or False
                        Lower Constraint
                        Upper Constraint
                        Step Size
            
        '''
        optimization_variables_dict = {}
        opt_dict_idx = 0
        Documents_dict = {0:self.awrde.Project.Schematics,
                          1:self.awrde.Project.SystemDiagrams,
                          2:self.awrde.Project.GlobalDefinitionDocuments,
                          3:self.awrde.Project.EMStructures}
        DocType_dict = {0:'Circuit Schematic', 1:'System Diagram', 2:'Global Definitions',
                        3:'EM Structure'}
        NumDocuments = len(Documents_dict)
        #
        for doc_idx in range(NumDocuments):
            Documents = Documents_dict[doc_idx]
            for document in Documents:
                if DocType_dict[doc_idx] == 'Global Definitions':
                    Elements_object = document.DataElements
                    Equations_object = document.Equations
                elif DocType_dict[doc_idx] == 'EM Structure':
                    Elements_object = document.Schematic.Elements
                    Equations_object = document.Schematic.Equations
                else:
                    Elements_object = document.Elements
                    Equations_object = document.Equations
                #end if
                for element in Elements_object:
                    if element.Enabled:
                        for param in element.Parameters:
                            if param.Optimize:
                                Fields_dict = {}
                                Fields_dict['Document Name'] = document.Name
                                Fields_dict['Document Type'] = DocType_dict[doc_idx]
                                Fields_dict['Element Type'] = 'Element'
                                Fields_dict['Variable Name'] = element.Name + ' : '+param.Name
                                Fields_dict['Constrained'] = param.Constrain
                                Fields_dict['Nominal Value'] = param.ValueAsDouble
                                Fields_dict['Lower Constraint'] = param.LowerConstraint
                                Fields_dict['Upper Constraint'] = param.UpperConstraint
                                Fields_dict['Step Size'] = param.StepSize
                                optimization_variables_dict[opt_dict_idx] = Fields_dict
                                opt_dict_idx += 1
                            #end if
                        #end for
                    #end if
                #end for
                for eqn in Equations_object:
                    if eqn.Enabled:
                        if eqn.Optimize:
                            Fields_dict = {}
                            Fields_dict['Document Name'] = document.Name
                            Fields_dict['Document Type'] = DocType_dict[doc_idx]
                            Fields_dict['Element Type'] = 'Equation'
                            Expression = eqn.Expression
                            Expression = Expression.lstrip()
                            Expression = Expression.rstrip()
                            Fields_dict['Variable Name'] = Expression
                            Fields_dict['Constrained'] = eqn.Constrain
                            try:
                                val_str = eqn.Expression.split('=')
                                val = float(val_str[1])
                            except:
                                val = 'NA'
                            #end try
                            Fields_dict['Nominal Value'] = val
                            Fields_dict['Lower Constraint'] = eqn.LowerConstraint
                            Fields_dict['Upper Constraint'] = eqn.UpperConstraint
                            Fields_dict['Step Size'] = eqn.StepSize
                            optimization_variables_dict[opt_dict_idx] = Fields_dict
                            opt_dict_idx += 1
                        #end if
                    #end if
                #end for
            #end for
        #end for
        return optimization_variables_dict
        #
    def print_optimization_variables(self, optimization_variables_dict:dict):#----------------
        '''prints the optimization variables and their attributes'''
        NumOptVars = len(optimization_variables_dict)
        ov = optimization_variables_dict
        for opt_idx in range(NumOptVars):
            print('{:s}{:3.0f}'.format('Optimization Variable Index =',opt_idx))
            print('    {:s} {:s}'.format('Document Name =',ov[opt_idx]['Document Name']))
            print('    {:s} {:s}'.format('Document Type =',ov[opt_idx]['Document Type']))
            print('    {:s} {:s}'.format('Element Type =',ov[opt_idx]['Element Type']))
            if ov[opt_idx]['Element Type'] == 'Element':
                print('    {:s} {:s}'.format('Element Name : Parameter Name =',ov[opt_idx]['Variable Name']))
            elif ov[opt_idx]['Element Type'] == 'Equation':
                print('    {:s}{:s}{:s}'.format('Expression = "',ov[opt_idx]['Variable Name'],'"'))
            #end if
            print('    {:s} {:s}'.format('Constrained =',str(ov[opt_idx]['Constrained'])))
            print('    {:s} {:s}'.format('Nominal Value =',str(ov[opt_idx]['Nominal Value'])))
            print('    {:s} {:s}'.format('Lower Constraint =',str(ov[opt_idx]['Lower Constraint'])))
            print('    {:s} {:s}'.format('Upper Constraint =',str(ov[opt_idx]['Upper Constraint'])))
            print('    {:s} {:s}'.format('Step Size =',str(ov[opt_idx]['Step Size'])))
            print('')
        #end for
        #
    @property
    def optimization_goals_dict(self) -> dict:#------------------------------------------
        '''
        returns dictionary of optimizer goals
        
        Returns
        -------
        optimization_goals_dict: dictionary
             Keys: Index
             Values: Dictionary of fields
                        
        
        '''
        self._build_opt_goal_dict()
        return self._optimization_goals_dict
        #
    def add_optimization_goal(self, measurement_name:str, goal_type:str, goal_xrange:tuple,
                              goal_yrange:tuple, goal_weight:float, goal_l_factor:float):
        '''
        add optimization goal to project
        
        Parameters
        ----------
           measurement_name : string   The measurement name in form of doc name : measurmeent
           goal_type : string  '>', '<', or '='
           Xrage : tuple  (xStart, xStop)  must be in base units
           Yrange : tuple (yStart, yStop)  
           
        '''
        if not isinstance(measurement_name, str):
            raise TypeError('measurement_name must be a string')
        if not isinstance(goal_type, str):
            raise TypeError('goal_type must be a string')
        if not isinstance(goal_yrange, tuple):
            raise TypeError('goal_xrange must be a tuple')
        if not isinstance(goal_yrange,tuple):
            raise TypeError('goal_yrange must be a tuple')
        if (not isinstance(goal_weight, float)) and (not isinstance(goal_weight, int)):
            raise TypeError('goal_weight must be either float or int type')        
        if (not isinstance(goal_l_factor, float)) and (not isinstance(goal_l_factor, int)):
            raise TypeError('goal_weight must be either float or int type')        
        #end if
        #
        if goal_type not in {'<','>','='}:
            raise RuntimeError('invalid goal_type: ' + goal_type)
        #end if
        #
        meas_name_split = measurement_name.split(':')
        CircuitName = meas_name_split[0]
        MeasName = meas_name_split[1]
        if goal_type == '=':
            GoalTypeEnum = 0
        elif goal_type == '<':
            GoalTypeEnum = 1
        elif goal_type == '>':
            GoalTypeEnum = 2
        else:
            raise RuntimeError('invalid goal_type: ' + goal_type)
        #end if
        xStart = goal_xrange[0]
        xStop = goal_xrange[1]
        yStart = goal_yrange[0]
        yStop = goal_yrange[1]
        xUnit = 0
        yUnit = 0
        #
        try:
            goal_idx = self.awrde.Project.OptGoals.Count
            self.awrde.Project.OptGoals.AddGoal(CircuitName, MeasName, GoalTypeEnum, goal_weight,
                                        goal_l_factor, xStart, xStop, xUnit, yStart, yStop, yUnit)
            self._optimization_goals_dict[goal_idx] = _Goal(self.awrde, goal_idx, 'Optimize')
        except:
            warnings.warn('optimization goal could not be added')
        #end try
        #
    def remove_optimization_goal(self, optimization_goal_name:str):#------------------------------
        '''
        Remove optimiztion goal
        '''
        if not isinstance(optimization_goal_name, str):
            raise TypeError('optimization_goal_name must be string type')
        #end if
        opt_goal_removed = False
        #
        for i in range(len(self._optimization_goals_dict)):
            temp_opt_goal = self._optimization_goals_dict[i]
            if optimization_goal_name == temp_opt_goal.goal_name:
                dict_idx = i
                break
            #end if
        #end for
        #
        for opt_goal_idx in range(1, self.awrde.Project.OptGoals.Count+1):
            temp_opt_goal_name = self.awrde.Project.OptGoals(opt_goal_idx).Name
            if temp_opt_goal_name == optimization_goal_name:
                self.awrde.Project.OptGoals.Remove(opt_goal_idx)
                opt_goal_removed = True
                self._optimization_goals_dict.pop(dict_idx)
                break
            #end if
        #end for
        if not opt_goal_removed:
            warnings.warn('Optimization Goal ' + optimization_goal_name + ' did not get removed')
        #end for
        return opt_goal_removed
        #
    @property
    def total_opt_goal_cost(self) -> float:#---------------------------------------------------
        '''returns total cost of the all the combined goals'''
        return self.awrde.Project.OptGoals.TotalCost
        #
    @property
    def optimization_type_list(self) -> list:#-----------------------------------------------
        '''
        returns list of all the avialble optimization methods
        
        '''
        opt_type_list = []
        for i in range(1, self.awrde.Project.Optimizer.TypeCount+1):
            opt_type_list.append(self.awrde.Project.Optimizer.TypeName(i))
        #end for
        return opt_type_list
        #
    @property
    def optimization_type(self) -> str: #-----------------------------------------------
        '''
        returns optimization type name
        '''
        TypeNum = self.awrde.Project.Optimizer.Type
        self._build_opt_properties_dict()  #build the properties dictionary upon querry of the optimization type
        return self.awrde.Project.Optimizer.TypeName(TypeNum)
        #
    @optimization_type.setter
    def optimization_type(self, optimzation_type_name: str):#---------------------------
        '''sets the current optimization type'''
        if not isinstance(optimzation_type_name, str):
            raise TypeError('optimization_type_name must be a string')
        #end if
        found_it = False
        for i in range(1, self.awrde.Project.Optimizer.TypeCount+1):
            if optimzation_type_name == self.awrde.Project.Optimizer.TypeName(i):
                found_it = True
                break
            #end if
        #end for
        if not found_it:
            raise RuntimeError('Invalid optimization type: '+optimzation_type_name)
        #end if
        self.awrde.Project.Optimizer.Type = i
        self._build_opt_properties_dict() #Update the optimization properties dictionary upon changing the optimization type
        #
    @property
    def optimization_type_properties(self) -> dict:#-------------------------------------------
        '''
        returns a dictionary of properties for the current optimization type
        
        Returns
        -------
        opt_properties_dict: dictionary
                             keys are the property name
                             values are the property values
        '''
        try:
            opt_properties_dict = self._opt_properties_dict
        except:
            self._build_opt_properties_dict()
            opt_properties_dict = self._opt_properties_dict
        #end try
        return opt_properties_dict
        #
    def optimization_update_type_properties(self):#----------------------------------------------
        '''update the optimization type properties'''
        try:
            for prop in self.awrde.Project.Optimizer.Properties:
                prop.Value = self._opt_properties_dict[prop.Name]
            #end for
        except:
            warnings.warn('Optimization properties not updated')
        #end try
        #
    @property
    def optimization_max_iterations(self) -> float:#------------------------------------------
        ''' returns the Optimizer MaxIterations'''
        return self.awrde.Project.Optimizer.MaxIterations
        #
    @optimization_max_iterations.setter
    def optimization_max_iterations(self, max_iterations: float):#---------------------------
        '''sets the Optimizer MaxIterations'''
        if (not isinstance(max_iterations, float)) and (not isinstance(max_iterations, int)):
            raise TypeError('max_iterations must be either a float or int type')
        #end if
        #
        try:
            self.awrde.Project.Optimizer.MaxIterations = max_iterations
        except:
            warnings.warn('Optimizer max iterations could not be set')
        #end try
        self.awrde.Project.Optimizer.MaxIterations = int(max_iterations)
        #
    @property
    def optimization_show_all_iterations(self) -> bool:#---------------------------------------------
        return self.awrde.Project.Optimizer.ShowAllUpdates
        #
    @optimization_show_all_iterations.setter
    def optimization_show_all_iterations(self, show_all_iterations: bool):#----------------------
        if not isinstance(show_all_iterations, bool):
            raise TypeError('show_all_iterations must be boolean type')
        #end if
        self.awrde.Project.Optimizer.ShowAllUpdates = show_all_iterations
        #
    @property
    def optimization_stop_at_minimum_error(self) -> bool:#---------------------------------------------
        return self.awrde.Project.Optimizer.StopAtMin
        #
    @optimization_stop_at_minimum_error.setter
    def optimization_stop_at_minimum_error(self, stop_at_min_error: bool):#----------------------
        if not isinstance(stop_at_min_error, bool):
            raise TypeError('stop_at_min_error must be boolean type')
        #end if
        self.awrde.Project.Optimizer.StopAtMin = stop_at_min_error
        #
    @property
    def optimization_stop_on_simulation_error(self) -> bool:#---------------------------------------------
        return self.awrde.Project.Optimizer.StopOnErr
        #
    @optimization_stop_on_simulation_error.setter
    def optimization_stop_on_simulation_error(self, stop_on_sim_error: bool):#----------------------
        if not isinstance(stop_on_sim_error, bool):
            raise TypeError('stop_on_sim_error must be boolean type')
        #end if
        self.awrde.Project.Optimizer.StopOnErr = stop_on_sim_error
        #
    @property
    def optimization_cancel_on_stop_request(self) -> bool:#---------------------------------------------
        return self.awrde.Project.Optimizer.CancelOnStop
        #
    @optimization_cancel_on_stop_request.setter
    def optimization_cancel_on_stop_request(self, cancel_on_stop_request: bool):#----------------------
        if not isinstance(cancel_on_stop_request, bool):
            raise TypeError('cancel_on_stop_request must be boolean type')
        #end if
        self.awrde.Project.Optimizer.CancelOnStop = cancel_on_stop_request
        #
    @property
    def optimization_log_to_file(self) -> bool:#---------------------------------------------
        return self.awrde.Project.Optimizer.LogToFile
        #
    @optimization_log_to_file.setter
    def optimization_log_to_file(self, log_to_file: bool):#----------------------
        if not isinstance(log_to_file, bool):
            raise TypeError('log_to_file must be boolean type')
        #end if
        self.awrde.Project.Optimizer.LogToFile = log_to_file
        #    
    @property
    def optimization_cost(self) -> float:#-----------------------------------------------------
        return self.awrde.Project.Optimizer.Cost
        #
    @property
    def optimization_best_cost(self) -> float:#-----------------------------------------------------
        return self.awrde.Project.Optimizer.BestCost
        #
    @property
    def optimization_current_iteration(self) -> int:#---------------------------------------------
        return self.awrde.Project.Iterataion
        #
    @property
    def optimization_start(self) -> bool:#---------------------------------------------------------
        return self.awrde.Project.Optimizer.Running
        #
    @optimization_start.setter
    def optimization_start(self, opt_start: bool):#----------------------------------------
        if not isinstance(opt_start, bool):
            raise TypeError('opt_start must be boolean type')
        #end if
        if opt_start:
            self.awrde.Project.Optimizer.start()  #yes, lower case start()
        else:
            self.awrde.Project.Optimizer.stop()  #yes, lower case stop()
        #end if
        #
    def optimization_reset(self):#----------------------------------------------------------
        self.awrde.Project.Optimizer.Reset()
        #
    def optimization_round(self, num_significant_digits: int):#---------------------------------
        if not isinstance(num_significant_digits, int):
            raise TypeError('num_significant_digits must be int type')
        #end if
        self.awrde.Project.Optimizer.Round(num_significant_digits, True, True)
        #
    @property
    def optimization_log_file_select(self) -> str:#------------------------------------------
        '''
        Open up file brower pointed to the log file directory. User selects 
        desired optimization log file. These are .json formatted files
        
        Returns
        -------
             Dir_n_File: string.  Full path to the selected optimization .json file
        '''
        from tkinter import filedialog
        NumDirectories = self.awrde.Project.Application.Directories.Count
        for dir_idx in range(1, NumDirectories+1):
            DirName = self.awrde.Project.Application.Directories.Item(dir_idx).Name
            if DirName == 'Logs':
                LogDirPath = self.awrde.Project.Application.Directories.Item(dir_idx).ValueAsString
            #end if
        #end for
        filetypes = (('json files','*.json'),('All files', '*.*'))
        Dir_n_File = filedialog.askopenfilename(title='Select .json file', initialdir=LogDirPath, filetypes=filetypes)
        return Dir_n_File
        #
    def optimization_extract_log_file(self, Dir_n_File: str) -> object:#---------------------------
        '''returns JSON formated optimization log file'''        
        try:
            json_file_object = open(Dir_n_File)
            OptLogFile = json.load(json_file_object)
        except:
            raise RuntimeError('Could not extract optimzation log file')
        #end try
        return OptLogFile        
        #
    def optimization_log_file_params_dict(self, OptLogFile: object) -> dict:#--------------------------------
        '''
        Returns dictionay of parameters extracted from optimization log file
        
        Parameters
        ----------
        OptLogFile: JSON formated optimization log file as read by optimization_extract_log_file()
        
        Returns
        -------
        OptParam_dict: dictionary
                       Keys:
                          Optimization Type
                          MaxIterations
                          NumIterations
                          Properties
                          Stop at min error
                          Stop on simulation error
                          Cancel on stop request
                          Show all iterations
        '''
        OptParam_dict = {}
        try:
            OptParam_dict['Optimization Type'] = OptLogFile['Setup']['optimizer']['name']
            OptParam_dict['MaxIterations'] = OptLogFile['Setup']['optimizer']['maxiter']
            OptParam_dict['NumIterations'] = len(OptLogFile['Iterations'])
            OptParam_dict['Properties'] = OptLogFile['Setup']['optimizer']['props']
            OptParam_dict['Stop at min error'] = OptLogFile['Setup']['optimizer']['stopatmin']
            OptParam_dict['Stop on simulation error'] = OptLogFile['Setup']['optimizer']['stoponerr']
            OptParam_dict['Cancel on stop request'] = OptLogFile['Setup']['optimizer']['cancelonstop']
            OptParam_dict['Show all iterations'] = OptLogFile['Setup']['optimizer']['showallupdates']
        except:
            raise RuntimeError('Could not retrieve paramters from optimization log file')
        #end try
        return OptParam_dict
        #
    def optimization_log_file_goals_dict(self, OptLogFile: object) -> dict:#--------------------------------------
        '''
        Returns dictionary of optimization goals read from the optimization log file
        
        Parameters
        ----------
        OptLogFile: JSON formated optimization log file as read by optimization_extract_log_file()
        
        Returns
        -------
        OptGoal_dict: dictionary
                      Keys: Goal Name
                      Values: dictionary
                              Keys: Type, Meas, Enaabled, Weight, L, Xstart, Ystart,
                                    Xstop, Ystop, Tag
        '''
        OptGoal_dict = {}
        try:
            for goal_name, goal_dict in OptLogFile['Setup']['goals'].items():
                OptGoal_dict[goal_name] = goal_dict
            #end for
        except:
            raise RuntimeError('Could not retrieve goals from optimization log file')
        #end try
        return OptGoal_dict
        #
    def optimization_log_file_opt_variables_dict(self, OptLogFile: object) -> dict:#-----------------
        '''
        Returns dictionary of optimization variables read from the optimization
        log file
        
        Parameters
        ----------
        OptLogFile: JSON formated optimization log file as read by optimization_extract_log_file()
        
        Retruns
        -------
        OptVariables_dict: dictionary
        
        '''
        OptVariables_dict = {}
        try:
            IterationsKeys_list = list(OptLogFile['Iterations'].keys())
            LastIteration_dict = OptLogFile['Iterations'][IterationsKeys_list[-1]]
            FinalValues_list = LastIteration_dict['params']
            i = 0
            for var_name, var_value_list in OptLogFile['Setup']['vars'].items():
                Temp_dict = {}
                Temp_dict['Initial Value'] = var_value_list[0]
                Temp_dict['Min Constrain Value'] = var_value_list[1]
                Temp_dict['Max Constrain Value'] = var_value_list[2]
                Temp_dict['Step Size'] = var_value_list[3]
                Temp_dict['Final Value'] = FinalValues_list[i]
                OptVariables_dict[var_name] = Temp_dict
                i += 1
            #end for
        except:
            raise RuntimeError('Could not retrieve optimization variables from optimization log file')
        #end try
        return OptVariables_dict
        #
    def optimization_log_file_cost_array(self, OptLogFile: object):#--------------------------------
        '''
        Returns array of costs for each optimization iteration. Read from optimization
        log file.
        
        Parameters
        ----------
        OptLogFile: JSON formated optimization log file as read by optimization_extract_log_file()
        
        Returns
        -------
        TotalCost_ay: array
        
        '''
        TotalCost_list = list()
        try:
            keys_list = OptLogFile['Iterations'].keys()
            for key in keys_list:
                TotalCost_list.append(OptLogFile['Iterations'][key]['totalcost'])
            #end for
        except:
            print('Could not retrive cost array from optimization log file')
        #end try
        TotalCost_ay = np.array(TotalCost_list)
        return TotalCost_ay
        #
    def _build_opt_goal_dict(self) -> dict:#---------------------------------------------------
        self._optimization_goals_dict = {}
        for goal_idx in range(self.awrde.Project.OptGoals.Count):
            self._optimization_goals_dict[goal_idx] = _Goal(self.awrde, goal_idx, 'Optimize')
        #end for
        #        
    def _build_opt_properties_dict(self) -> dict:#--------------------------------------------
        ''' builds a dictionary of optimization properties '''
        self._opt_properties_dict = {}
        for prop in self.awrde.Project.Optimizer.Properties:
            self._opt_properties_dict[prop.Name] = prop.Value
        #end for
        #
#
#*****************************************************************************
#
class _GlobalDefinitions():
    '''
    Methods for collection of Global Definitions Documents
    
     Parameters
    ----------
    awrde: The AWRDE object returned from awrde_utils.establish_link()
    '''
    def __init__(self, awrde):#----------------------------------------
        self.awrde = awrde
        #
    def _build_global_defs_dict(self):#--------------------------------
        '''Create list of all graph names in project
    
        Parameters
        ----------
        None
    
        Returns
        -------
        _gloabal_defs_dict: dict
                            keys are the names of the global definitions documents
                            values are the global definition document objects
        '''
        self._global_defs_dict = {}
        for gd in self.awrde.Project.GlobalDefinitionDocuments:
            global_def = _GlobalDefinition(self.awrde, gd.Name)
            self._global_defs_dict[gd.Name] = global_def
        #end for
        #
    @property
    def global_definitions_dict(self) -> dict:#-------------------------------------
        '''returns global definition documents dictionary'''
        self._build_global_defs_dict()
        return self._global_defs_dict
        #
    @property
    def global_definitions_list(self) -> list:#-----------------------------------
        '''returns global definition documents list of names'''
        gloabal_defs_list = list()
        for gd in self.awrde.Project.GlobalDefinitionDocuments:
            gloabal_defs_list.append(gd.Name)
        #end for
        return gloabal_defs_list
        #
    def add_global_definitions_document(self, global_def_doc_name: str):#-----------------------------------
        gd_doc_exists = False
        for gd in self.awrde.Project.GlobalDefinitionDocuments:
            if gd.Name == global_def_doc_name:
                gd_doc_exists = True
            #end if
        #end for
        if gd_doc_exists:
            YesNo = messagebox.askyesno('Add Global Definition Document','Global Definition Document ' + global_def_doc_name + ' exists. Delete existing Global Definion Document?')
            if YesNo:
                self.remove_global_definitions_document(global_def_doc_name)
            else:
                global_def_doc_name += ' 1'
            #end if
        #end if
        try:
            self.awrde.Project.GlobalDefinitionDocuments.Add(global_def_doc_name)
        except:
            warnings.warn('Global definition doc not added. Possible name conflict')
        #end try
        #
    def remove_global_definitions_document(self, global_def_doc_name: str) -> bool:#-----------------
        '''
        Delete a global definition document from the AWRDE project
        
        Parameters
        ----------
        global_def_doc_name: string,
                 name of the Global Definition Document  to be deleted
        
        Returns
        -------
        global_def_doc_removed: bool
                    True if global definition document successfully deleted,
                    False if global definition document could not be deleted
                    
        '''             
        if not isinstance(global_def_doc_name, str):
            raise TypeError('global_def_doc_name must be a string type')
        #end if
        global_def_doc_removed = False
        try:
            for gd_idx in range(self.awrde.Project.GlobalDefinitionDocuments.Count):
                print(self.awrde.Project.GlobalDefinitionDocuments[gd_idx].Name,'    ',global_def_doc_name)
                if self.awrde.Project.GlobalDefinitionDocuments[gd_idx].Name == global_def_doc_name:
                    self.awrde.Project.GlobalDefinitionDocuments.Remove(gd_idx+1)
                    global_def_doc_removed = True
                    break
                #end if
            #end for
        except:
            warnings.warn('remove_global_definitions_document: Global Definitions Document ' + global_def_doc_name + ' did not get removed')
        #end try
        return global_def_doc_removed
        #
#
#*****************************************************************************
#
class _GlobalDefinition(_Elements, _Equations):
    def __init__(self, awrde, global_def_name: str):#-----------------------------------
        self.awrde = awrde
        self._initialize_data_file(global_def_name)
        #
    def _initialize_data_file(self, global_def_name: str):#-------------------------------------
        try:
            self._gd = self.awrde.Project.GlobalDefinitionDocuments(global_def_name)
        except:
            raise RuntimeError('Global Definition Document does not exist: ' + global_def_name)
        #end try
        _Elements.__init__(self, self._gd, 'Global Definition')
        _Equations.__init__(self, self._gd)        
        #
    @property
    def global_def_name(self) -> str:#------------------------------------------------------
        '''Returns name of the Global Defintion Document'''
        return self._gd.Name
        #    
#
#*****************************************************************************
#


from pyawr_utils import common_utils
import warnings
from tkinter import messagebox
import numpy as np

_Elements = common_utils._Elements
_Equations = common_utils._Equations
_Document_Methods = common_utils._Document_Methods
#
#**********************************************************************************************
#
class _SystemDiagrams():
    def __init__(self, awrde):#----------------------------------------------------
        self.awrde = awrde
        self._build_sys_diags_dict()
        #
    def _build_sys_diags_dict(self):
        '''Create list of all  system diagrams names in project '''
        self._sys_diags_dict = {}
        for sys_diag in self.awrde.Project.SystemDiagrams:
            self._sys_diags_dict[sys_diag.Name] = _SystemDiagram(self.awrde, sys_diag.Name)
        #end for
        #
    @property
    def system_diagrams_dict(self) -> dict:#------------------------------------------
        '''returns dictionary of system diagrams'''
        self._build_sys_diags_dict()
        return self._sys_diags_dict
        #  
    @property
    def system_diagram_name_list(self) -> list[str]:#-------------------------------------
        '''
        Create list of all system diagram names in project
        
        Parameters
        ----------
        None
        
        Returns
        -------
        system_diagram_name_list: list[string]
                         Each element in the list is a system digram name
        '''        
        _system_diagram_name_list = list()
        for system_diagram in self.awrde.Project.SystemDiagrams:
            _system_diagram_name_list.append(system_diagram.Name)
        #end for
        return _system_diagram_name_list
        #
    def remove_system_diagram(self, system_diagram_name: str) -> bool:
        '''
        Delete a system diagram from the AWRDE project
        
        Parameters
        ----------
        system_diagram_name: string,
                 name of the System Diagram to be deleted
        
        Returns
        -------
        system_diagram_removed: bool
                    True if system diagram successfully deleted
                    False if system diagram could not be deleted
                    
        '''  
        if not isinstance(system_diagram_name, str):
            raise RuntimeError('system_diagram_name must be a string type')
        #end if
        system_diagram_removed = False
        try:
            for sys_idx in range(self.awrde.Project.SystemDiagrams.Count):
                if self.awrde.Project.SystemDiagrams[sys_idx].Name == system_diagram_name:
                    self.awrde.Project.SystemDiagrams.Remove(sys_idx+1)
                    system_diagram_removed = True
                    break
                #end if
            #end for
        except:
            warnings.warn('remove_system_diagram: System Diagram ' + system_diagram_name + ' did not get removed')
        #end for
        return system_diagram_removed
        #
    def rename_system_diagram(self, system_diagram_name: str, new_system_diagram_name: str) -> bool:
        '''
        rename a system diagram
        
        Parameters
        ----------
        system_diagram_name: string
        new_system_diagram_name: string
        
        '''
        if not isinstance(system_diagram_name, str):
            raise RuntimeError('system_diagram_name must be a string type')
        #end if
        if not isinstance(new_system_diagram_name, str):
            raise RuntimeError('new_system_diagram_name must be a string type')
        #end if
        sys_diagram_name_list = self.system_diagram_name_list
        found_it = False
        for sys_diag_name_from_list in sys_diagram_name_list:
            if sys_diag_name_from_list == system_diagram_name:
                found_it = True
                if new_system_diagram_name in sys_diagram_name_list:
                    YesNo = messagebox.askyesno('Rename System Diagram','System Diagram '\
                        + new_system_diagram_name + ' exists. Remove existing System Diagram ?')
                    if YesNo:
                        self.remove_system_diagram(new_system_diagram_name)
                    else:
                        new_system_diagram_name += ' 1'
                    #end if
                #end if
                try:
                    sys = self.awrde.Project.SystemDiagrams(system_diagram_name)
                    sys.Name = new_system_diagram_name
                except:
                    raise RuntimeError('problem with renaming system diagram')
                #end try
                break
        #end for 
        if not found_it:
            warnings.warn('rename system diagram failed: ' + system_diagram_name)
        #end if
        #
    def copy_system_diagram(self, system_diagram_name: str, new_system_diagram_name: str) -> bool:    
        '''
        copy a system diagram
        
        Parameters
        ----------
        system_diagram_name: string
                            existing system diagram name
        new_system_diagram_name: string
                            name of the new system diagram
        
        '''        
        if not isinstance(system_diagram_name, str):
            raise RuntimeError('system_diagram_name must be a string type')
        #end if
        if not isinstance(new_system_diagram_name, str):
            raise RuntimeError('new_system_diagram_name must be a string type')
        #end if
        sys_diagram_name_list = self.system_diagram_name_list
        found_it = False
        sys_diag_idx = 1
        for sys_diag_name in sys_diagram_name_list:
            if system_diagram_name == sys_diag_name:
                found_it = True
                try:
                    self.awrde.Project.SystemDiagrams.Copy(sys_diag_idx, new_system_diagram_name)
                except:
                    raise RuntimeError('problem with copying system diagram')
                #end try
                break
                #
            #end if
            sys_diag_idx += 1
        #end for
        if not found_it:
            warnings.warn('copy system diagram failed: ' + system_diagram_name)
        #end if        
        #
    @property
    def system_frequencies(self) -> np.ndarray: #-------------------------------------------
        '''returns system frequecies from project defaults'''
        system_freq_list = list()
        for freq in self.awrde.Project.SystemDiagrams.RFBFrequencyOffsets:
            system_freq_list.append(freq.Value)
        #end for
        system_freq_ay = np.round(np.array(system_freq_list),1)
        return system_freq_ay
        #
    def set_system_frequencies(self, system_freq_ay: np.ndarray, units_str: str = 'Hz'):#-----------------------
        '''sets project frequencies'''
        #Cannot be made as a setter property due to multiple pass parameters
        units_dict = {'Hz':1, 'kHz':1e3, 'MHz':1e6, 'GHz':1e9}
        try:
            freq_multiplier = units_dict[units_str]
        except:
            raise RuntimeError('Incorrect units: ' + units_str)
        #end try
        system_freq_ay *= freq_multiplier
        try:
            self.awrde.Project.SystemDiagrams.RFBFrequencyOffsets.Clear()
            self.awrde.Project.SystemDiagrams.RFBFrequencyOffsets.AddMultiple(system_freq_ay)
        except:
            raise RuntimeError('Error in setting system frequencies')
        #end try
        #
    @property
    def system_options_dict(self) -> dict: #-----------------------------------------------
        '''returns dictionary of system default options'''
        system_options = self.awrde.Project.SystemDiagrams.Options
        sys_options_dict = {}
        for opt in system_options:
            if opt.Name != '':  #One of the options has blank as the name
                sys_options_dict[opt.Name] = opt.Value
            #end if
        #end for
        return sys_options_dict
        #
    def set_system_options_dict(self, system_options_dict: dict):#-------------------------------
        '''set system default options'''
        for opt_name, opt_value in system_options_dict.items():
            try:
                self.awrde.Project.SystemDiagrams.Options(opt_name).Value = opt_value
            except:
                raise RuntimeError('setting default system options failed: ' + opt_name + ' , ' + str(opt_value))
            #end try
        #end for
        #
    @property  
    def read_lpf_name(self) -> str:#-------------------------------------------------
        '''
        All system diagrams use the same layout process file which is the first 
        
        Returns
        -------
        sys_diag_lpf: string
        '''
        sys_diag_lpf = self.awrde.Project.ProcessDefinitions[0].Name
        return sys_diag_lpf
        #
#
#**********************************************************************************************
#   
class _SystemDiagram(_Elements, _Equations, _Document_Methods):
    def __init__(self, awrde, system_diagram_name: str):#-----------------------------------------
        self.awrde = awrde
        self._initialize_system_diagram(system_diagram_name)
        #
    def _initialize_system_diagram(self, system_diagram_name):#--------------------------------------------------
        try:
            self._sys_diag = self.awrde.Project.SystemDiagrams(system_diagram_name)
        except:
            raise RuntimeError('System Diagram does not exist: ' + system_diagram_name)
        #end try
        #
        try:
            self._elem = self._sys_diag.Elements[0]
        except:
            self._elem = None
        #end try
        self._system_diagram_name = system_diagram_name
        _Elements.__init__(self, self._sys_diag, 'System')
        _Equations.__init__(self, self._sys_diag)
        _Document_Methods.__init__(self, self._sys_diag)
        #
    @property
    def system_diagram_name(self) -> str:#-----------------------------------------------------
        '''Returns name of System Diagram'''
        return self._system_diagram_name
        # 
    @property
    def frequencies(self) -> np.ndarray: #-------------------------------------------
        '''returns system frequecies from project defaults'''
        freq_list = list()
        for freq in self._sys_diag.RFBFrequencyOffsets:
            freq_list.append(freq.Value)
        #end for
        freq_array = np.round(np.array(freq_list),1)
        return freq_array
        #
    def set_frequencies(self, freq_array: np.ndarray, units_str: str = 'GHz'):#-----------------------
        '''sets project frequencies'''
        #Cannot be made as a setter property due to multiple pass parameters
        units_dict = {'Hz':1, 'kHz':1e3, 'MHz':1e6, 'GHz':1e9}
        try:
            freq_multiplier = units_dict[units_str]
        except:
            raise RuntimeError('Incorrect units: ' + units_str)
        #end try
        freq_array *= freq_multiplier
        try:
            self._sys_diag.RFBFrequencyOffsets.Clear()
            self._sys_diag.RFBFrequencyOffsets.AddMultiple(freq_array)
        except:
            raise RuntimeError('Error in setting system frequencies')
        #end try
        #
    @property
    def use_project_frequencies(self) -> bool:#----------------------------------------------
        return self._sys_diag.UseProjectRFBFrequencyOffsets
        #
    def set_use_project_frequencies(self, set_state: bool):#-------------------------------------
        if not isinstance(set_state, bool):
            raise TypeError('set_state must be a boolean')
        #end if
        self._sys_diag.UseProjectRFBFrequencyOffsets = set_state
        #
    @property
    def option_sets(self) -> dict:#----------------------------------------------------------
        '''returns dictionary of system diagram option sets'''
        sys_diagram_opt_set_dict = {}
        for optset in self._sys_diag.OptionSets:
            optset_dict = {}
            optset_name = optset.Name
            for optsubset in optset.SubSets:
                optsubset_name = optsubset.Name
                opt_dict = {}
                for opt in optsubset.Options:
                    opt_dict[opt.Name] = opt.Value
                #end for
                optset_dict[optsubset_name] = opt_dict
            #end for
            sys_diagram_opt_set_dict[optset_name] = optset_dict
        #end for
        return sys_diagram_opt_set_dict
        #
    def set_option_sets(self, sys_diagram_opt_set_dict: dict):#------------------------------
        '''sets dictionary of system diagram option sets'''
        for optset_name, optset_dict in sys_diagram_opt_set_dict.items():
            optset = self._sys_diag.OptionSets(optset_name)
            for optsubset_name, optsubset_dict in optset_dict.items():
                optsubset = optset.SubSets(optsubset_name)
                for opt_name, opt_value in optsubset_dict.items():
                    optsubset.Options(opt_name).Value = opt_value
                #end for
            #end for
        #end for
        #
    @property
    def show_options_sets(self):#---------------------------
        '''displays system diagram option sets'''
        sys_diagram_opt_set_dict = self.option_sets
        print('System Diagram Options')
        print('-'*40)        
        for optset_name, optset in sys_diagram_opt_set_dict.items():
            print('{:s}'.format(optset_name))
            for optsubset_name, optsubset in optset.items():
                print('   {:s}'.format(optsubset_name))
                for opt_name, opt_value in optsubset.items():
                    try:
                        if not isinstance(opt_value, str):
                            opt_value = str(opt_value)
                        #end if
                        print('      {:s} : {:s}'.format(opt_name, opt_value))
                    except:
                        print(opt_name, opt_value, type(opt_value))
                        raise RuntimeError('oops!')
                    #end try
                #end for
            #end for
        #end for
        #
    @property
    def option_sets_use_project_defaults(self) -> dict:#------------------------------------
        '''returns dictionary of UseProjectDefaults for options in  a system diagram'''
        use_project_defaults_dict = {}
        for optset in self._sys_diag.OptionSets:
            optsubset_dict = {}
            for optsubset in optset.SubSets:
                optsubset_dict[optsubset.Name] = optsubset.UseProjectDefaults
            #end for
            use_project_defaults_dict[optset.Name] = optsubset_dict
        #end for
        return use_project_defaults_dict
        #
    def set_option_sets_use_project_defaults(self, use_project_defaults_dict: dict):#----------
        '''sets UseProjectDefaults in system diagram options'''
        try:
            for optset_name, optset_dict in use_project_defaults_dict.items():
                optset = self._sys_diag.OptionSets(optset_name)
                for optsubset_name, use_defaults in optset_dict.items():
                    optsubset = optset.SubSets(optsubset_name)
                    optsubset.UseProjectDefaults = use_defaults
                #end for
            #end for
        except:
            raise RuntimeError('Error in applying dictionary for set_option_sets_use_project_defaults')
        #end try
        #
    @property
    def show_option_sets_use_project_defaults(self):
        '''displays the UseProjectDefaults dictionary'''
        print('Use Project Defaults')
        print('-'*40)
        use_project_defaults_dict = self.option_sets_use_project_defaults
        for optset_name, optset in use_project_defaults_dict.items():    #Print out UseDefaultOptions
            print('{:s}'.format(optset_name))
            for optsubset_name, use_defaults in optset.items():
                print('   {:s} : {:s}'.format(optsubset_name, str(use_defaults)))
            #end for
        #end for
        #
        


import pyawr.mwoffice as mwo
import numpy as np
from tkinter import messagebox
import warnings
#
#*****************************************************************************
#
class _CommonTools():
    def __init__(self, awrde):#----------------------------------------------------------
        self.awrde = awrde
        #
    @property
    def get_all_active_measurements(self) -> list:#------------------------------------
        '''returns a list of all the active measurements from all the graphs in the project'''
        active_meas_list = list()
        for gr in self.awrde.Project.Graphs:
            for meas in gr.Measurements:
                active_meas_list.append(meas.Name)
            #end for
        #end for
        return active_meas_list
        #    
#
#*****************************************************************************
#
class _Document_Methods():
    def __init__(self, document: object,):#--------------------------------------------
        self._document = document
        #
    def add_wire(self, start_xy: tuple, end_xy: tuple):
        '''
        Adds a wire to a schematic or system diagram
        
        Parameters
        ---------- 
        start_xy: tuple
              Start x,y cooridnates of the wire segment
              
        end_xy: tuple
              End x,y cooridnates of the wire segment     
        '''
        try:
            self._document.Wires.Add(int(start_xy[0]), int(start_xy[1]), int(end_xy[0]), int(end_xy[1]))
        except:
            warnings.warn('Could not add wire to document')
        #end try
        #
    def remove_wire(self, start_xy: tuple, end_xy: tuple):
        '''
        Removes a wire from a schematic or system diagram
        
        Parameters
        ---------- 
        start_xy: tuple
              Start x,y cooridnates of the wire segment
              
        end_xy: tuple
              End x,y cooridnates of the wire segment     
        '''
        num_wires = self._document.Wires.Count
        wire_removed = False
        for wire_idx in range(num_wires):
            wire_start_x = self._document.Wires.Item(wire_idx+1).x1
            wire_start_y = self._document.Wires.Item(wire_idx+1).y1
            wire_end_x = self._document.Wires.Item(wire_idx+1).x2
            wire_end_y = self._document.Wires.Item(wire_idx+1).y2
            found_it = True
            if wire_start_x != start_xy[0]:
                found_it = False
            if wire_start_y != start_xy[1]:
                found_it = False
            if wire_end_x != end_xy[0]:
                found_it = False
            if wire_end_y != end_xy[1]:
                found_it = False
            #end if
            if found_it:
                self._document.Wires.Remove(wire_idx+1)
                wire_removed = True
                break
            #end if
        #end for
        if not wire_removed:
            warnings.warn('Wire could not be removed')
        #end if
        #
    @property
    def wire_segments_dict(self) -> dict:
        '''
        returns dictionary of wire segments
        
        Returns
        ---------- 
        wire_segment_dict:
             keys = wire segment index
             values = tuple (start x, start y, end x, end y)

        '''
        wire_segment_dict = dict()
        num_wires = self._document.Wires.Count
        for wire_idx in range(num_wires):
            wire_start_x = self._document.Wires.Item(wire_idx+1).x1
            wire_start_y = self._document.Wires.Item(wire_idx+1).y1
            wire_end_x = self._document.Wires.Item(wire_idx+1).x2
            wire_end_y = self._document.Wires.Item(wire_idx+1).y2
            wire_segment_dict[wire_idx] = (wire_start_x, wire_start_y, wire_end_x, wire_end_y)
        #end for
        return wire_segment_dict
        #
#
#*****************************************************************************
#
class _Parameter():
    def __init__(self, parameter, document_type):
        self._parameter = parameter
        self._document_type = document_type

    @property
    def name(self) -> str:
        return self._parameter.Name

    @property
    def value(self):#-------------------------------------------------------------
        if self._document_type == 'System':
            raise RuntimeError('Only value_str is valid for System Diagram element parameters')
        #end if
        return self._parameter.ValueAsDouble

    @value.setter
    def value(self, value):#-----------------------------------------------------
        if not (isinstance(value, float) or isinstance(value, int)):
            raise RuntimeError('value must be float or int type')
        #end if        
        if self._document_type == 'System':
            raise RuntimeError('Only value_str is valid for System Diagram element parameters')
        #end if
        self._parameter.ValueAsDouble = value
        #
    @property
    def value_str(self) -> str:#-----------------------------------------------------
        return self._parameter.ValueAsString
        #
    @value_str.setter
    def value_str(self, value_str: str):#-------------------------------------------
        if not isinstance(value_str, str):
            raise RuntimeError('value_str must be string type')
        #end if
        self._parameter.ValueAsString = value_str
        #
    @property
    def value_float(self) -> float:#--------------------------------------------------
        '''For system diagrams where value string cannot be used'''
        return self._parameter.ValueAsDouble
        #
    @value_float.setter
    def value_float(self, value_float: float):#------------------------------------------
        if not isinstance(value_float, float):
            raise TypeError('value_float must be float type')
        #end if
        self._parameter.ValueAsDouble = value_float
        #
    @property
    def optimize_enabled(self) -> bool:#----------------------------------------------
        '''returns Optimize enabled state '''
        return self._parameter.Optimize
        #
    @optimize_enabled.setter
    def optimize_enabled(self, enabled: bool):#--------------------------------------
        '''sets Optimize enabled state '''
        self._parameter.Optimize = enabled
        #
    @property
    def constrain(self) -> bool:#-------------------------------------------------------
        '''returns Contrain enabled state '''
        return self._parameter.Constrain
        #
    @constrain.setter
    def constrain(self, enabled: bool):#----------------------------------------------------
        '''sets Contrain enabled state '''
        self._parameter.Constrain = enabled
        #
    @property
    def lower_constraint(self) -> float:#-------------------------------------------------
        ''' returns LowerContraint value'''
        return self._parameter.LowerConstraint
        #
    @lower_constraint.setter
    def lower_constraint(self, lower_contraint_value:float):#-----------------------------
        ''' sets LowerContraint value'''
        if not isinstance(lower_contraint_value, float) and  not isinstance(lower_contraint_value, int):
            raise TypeError('lower_contraint_value must be float or int')
        #end if
        self._parameter.LowerConstraint = lower_contraint_value
        #
    @property
    def upper_constraint(self) -> float:#-------------------------------------------------
        ''' returns UpperContraint value'''
        return self._parameter.UpperConstraint
        #
    @upper_constraint.setter
    def upper_constraint(self, upper_contraint_value:float):#-----------------------------
        ''' sets UpperContraint value'''
        if not isinstance(upper_contraint_value, float) and  not isinstance(upper_contraint_value, int):
            raise TypeError('upper_contraint_value must be float or int')
        #end if
        self._parameter.UpperConstraint = upper_contraint_value
        #
    @property
    def step_size(self) -> float:#-------------------------------------------------
        ''' returns StepSize value'''
        return self._parameter.StepSize
        #
    @step_size.setter
    def step_size(self, step_size_value:float):#-----------------------------
        ''' sets StepSize value'''
        if not isinstance(step_size_value, float) and  not isinstance(step_size_value, int):
            raise TypeError('step_size_value must be float or int')
        #end if
        self._parameter.StepSize = step_size_value
        #    
class _Parameters():
    def __init__(self, element, document_type):#----------------------------------------------
        self._element = element
        self._document_type = document_type

    def _build_parameters_dict(self):
        self._parameters_dict = {}

        for parameter in self._element.Parameters:
            try:
                self._parameters_dict[parameter.Name] = _Parameter(self._element.Parameters(parameter.Name), self._document_type)
            except:
                raise RunetimeError('Failed to create dictionary of Parameters')
    
    @property
    def parameters_dict(self):
        '''
        I made this a getter with the @property decorator to keep this consistent with _Elements
        and _Equations. mainclean.py will error out on schem_LPF_C1_parameters = schem_LPF_C1.parameters_dict()
        '''
        self._build_parameters_dict()
        return self._parameters_dict
    
    @property
    def parameter_names_list(self) -> list[str]:#----------------------------------------------------
        '''
        creates a list of element names of all the elements in the document
        
        Parameters
        ----------
        None
        
        Returns
        -------
        param_list: list[string]
                   Each item in the list is parameter name
        '''        
        param_list = list()
        for parameter in self._element.Parameters:
            param_list.append(parameter.Name)

        return param_list

#
#**********************************************************************************************
#
class _Elements():
    def __init__(self, document, document_type):
        self._document = document
        self._document_type = document_type
        self._build_elements_dict()

    def _build_elements_dict(self):
        self._elements_dict = {}
        if self._document_type == 'Global Definition':
            for element in self._document.DataElements:
                try:
                    self._elements_dict[element.Name] = _Element(self._document.DataElements(element.Name), self._document_type)
                except:
                    raise RuntimeError('Failed to create dictionary of Elements')
                #end try
            #end for
        else:
            for element in self._document.Elements:
                try:
                    self._elements_dict[element.Name] = _Element(self._document.Elements(element.Name), self._document_type)
                except:
                    raise RuntimeError('Failed to create dictionary of Elements')
                #end try
            #end for
        #end if
        #
    @property
    def elements_dict(self) -> dict:
        return self._elements_dict
    
    @property
    def element_names_list(self) -> list[str]:#------------------------------------------
        '''
        creates a list of element names of all the elements in the document
        
        Parameters
        ----------
        None
        
        Returns
        -------
        elements_list: list[string]
                   Each item in the list is an element name
        '''
        elements_list = list()
        if self._document_type == 'Global Definition':
            for element in self._document.DataElements:
                elements_list.append(element.Name)
            #end for
        else:
            for element in self._document.Elements:
                elements_list.append(element.Name)
            #end for
        #end if
        return elements_list

    def add_element(self, element_name, x_pos: float, y_pos: float, rotation: float=0, flipped: bool=False):
        '''
        Add an element to the document
        
        Parameters
        ----------
        element_name: string,
                 name of the element to be added. 
        x_pos: float
        y_pos: float
        rotation: float, optional
                 default = 0
        flipped: bool, optional
                 default = False
                 
        Returns
        -------
        none
                    
        '''          
        if not isinstance(element_name, str):
            raise TypeError('element_name must be string type')
        if not (isinstance(x_pos, float) or isinstance(x_pos, int)):
            raise TypeError('x_pos must be float or int type')
        if not (isinstance(y_pos, float) or isinstance(y_pos, int)):
            raise TypeError('y_pos must be float or int type')
        if not (isinstance(rotation, float) or isinstance(rotation, int)):
            raise TypeError('rotation must be float or int type') 
        if not isinstance(flipped, bool):
            raise TypeError('flipped must be bool type')
        #end if
        #
        try:
            new_element = self._document.Elements.Add(element_name, x_pos, y_pos, rotation, flipped)
            self._elements_dict[new_element.Name] = _Element(self._document.Elements(new_element.Name), self._document_type)
        except:
            raise RuntimeError('Error when adding element: '+element_name)
        #end try
        #
    def remove_element(self, element_name_n_id: str) -> bool:#-----------------------------------------
        '''
        Delete an element from the document
        
        Parameters
        ----------
        element_name_n_id: string,
                 name of the element to be deleted. Must include ID
                 Format:  element_Name.element_ID. For instance: 'CAP.C1'
        
        Returns
        -------
        element_removed: bool
                    True if element successfully deleted
                    False if element could not be deleted
                    
        '''          
        if not isinstance(element_name_n_id, str):
            raise TypeError('element_name_n_id must be a string')
        #end if
        element_removed = False
        try:
            for ele_idx in range(self._document.Elements.Count):
                if self._document.Elements[ele_idx].Name == element_name_n_id:
                    self._document.Elements.Remove(ele_idx+1)
                    element_removed = True
                    self._elements_dict.pop(element_name_n_id)
                    break
                #end if
            #end for
        except:
            warnings.warn('remove_element: Element ' + element_name_n_id + ' did not get removed')
        #end try
        return element_removed
        #
#
#**********************************************************************************************
#
class _Element(_Parameters):
    def __init__(self, element, document_type):
        _Parameters.__init__(self, element, document_type)
        self._element = element
        self._document_type = document_type
        #
    def set_optimization(self, optimization_bounds: np.ndarray ) -> bool:
        pass
        #
    @property
    def element_name(self) -> str: #------------------------------------------
        '''returns element name'''
        return self._element.Name
        #    
    @property
    def xy_position(self) -> tuple:#-------------------------------------------------
        x_position = self._element.x
        y_position = self._element.y
        xy_position_tuple = (x_position, y_position)
        return xy_position_tuple
        #    
    def set_xy_position(self, xy_position_tuple: tuple):#------------------------------
        '''set xy position'''
        self._element.x = xy_position_tuple[0]
        self._element.y = xy_position_tuple[1]
        #
    @property
    def element_enabled(self) -> bool:#------------------------------------------------------       
        return self._element.Enabled
        #
    @element_enabled.setter
    def element_enabled(self,  enabled_state: bool):#-----------------------------------------
        if not isinstance(enabled_state, bool):
            raise TypeError('enabled_state must be boolean type')
        #end if        
        self._element.Enabled = enabled_state
        #
    @property
    def element_nodes_dict(self) -> dict:#--------------------------------------------------------
        ''' 
        returns dictionary of node x,y values
        
        Returns
        -------
             nodes_dict: dictionary 
                         dictionary keys are the node numbers
                         dictionary values are tuples with x,y values of the node
        '''
        nodes_dict = dict()
        num_nodes = self._element.Nodes.Count
        for node_idx in range(num_nodes):
            nodes_dict[node_idx] = (self._element.Nodes[node_idx].x, self._element.Nodes[node_idx].y)
        #end for
        return nodes_dict
        #
    
#
#**********************************************************************************************
#
class _Equations():
    '''
    methods for equations
    
    Parameters
    ----------    
    document: object
          schematic, system diagram or output equation documement object variable       
    '''
    def __init__(self, document):#----------------------------
        self._eqns_doc = document
        self._eqns_document_name = self._eqns_doc.Name
        self._build_equations_dict()
        #
    def _build_equations_dict(self):#--------------------------------------------------
        '''
        creates dictionary of equation names whose value is the variable type
        '''
        self._equations_dict = {}
        eqn_idx = 0
        for eqn in self._eqns_doc.Equations:
            equation = _Equation(eqn.Expression, self._eqns_doc)
            self._equations_dict[eqn_idx] = equation
            eqn_idx += 1
        #end for
        #
    @property
    def equations_dict(self) -> dict: #----------------------------------------------
        '''
        returns dictionary of equations. Key is the equaion name, vaue is the 
        equaiton value
        '''
        self._build_equations_dict()
        try:
            return self._equations_dict
        except:
            raise RuntimeError('equations_dict not available')
        #end try
        #
    @property
    def expression_list(self) -> list[str]:#-----------------------------------------
        '''
        Creates list of all expression in the document
        
        Parameters
        ----------
        None
        
        Returns
        -------
        expression_list: list[string]
                   Each item in the list is an expression in the document
        '''        
        expression_list = list()
        for eqn in self._eqns_doc.Equations:
            expression_list.append(eqn.Expression)
        #end for
        return expression_list
        #
    def remove_equation(self, expression: str):#----------------------------------
        '''
        Remove equation from system diagram
        
        Parameters
        ----------
        expression: string
        
        Returns
        -------
        equation_removed: boolean
        '''        
        if not isinstance(expression, str):
            raise TypeError('expression must be a string')
        #end if
        equation_removed = False
        for eqn_idx in range(1, self._eqns_doc.Equations.Count+1):
            eqn_expression = self._eqns_doc.Equations(eqn_idx).Expression
            if eqn_expression == expression:
                self._eqns_doc.Equations.Remove(eqn_idx)
                equation_removed = True
                break
            #end if
        #end for
        if not equation_removed:
            warnings.warn('equations did not get removed')
        #end if
        return equation_removed
        #
    def add_equation(self, equation_name: str, variable_type: str, equation_value, x_pos: float, y_pos: float):
        '''
        Add equation to system diagram
        
        Parameters
        ----------
        equation_name: string
        variable_type: string
                 Valid stings: Variable definition, Parameter definition, Display value
        equation_value: string, int or float
        x_pos: float
        y_pos: float
        '''        
        if not isinstance(equation_name, str):
            raise TypeError('equation_name must be a string')
        if not isinstance(variable_type, str):
            raise TypeError('variable_type must be a sting')
        #end if
        #
        if isinstance(equation_value, int) or isinstance(equation_value, float):
            equation_value = str(equation_value)
        elif not isinstance(equation_value, str):
            raise TypeError('equation_value type must be str, float or int')
        #end if        
        #
        equation_name = equation_name.lstrip() #Remove leading blanks
        equation_name = equation_name.rstrip() #Remove trialing blanks
        #
        equation_value = equation_value.lstrip() #Remove leading blanks
        equation_value = equation_value.rstrip() #Remove trialing blanks
        #
        variable_type = variable_type.lstrip()
        variable_type = variable_type.rstrip()
        variable_type_list = ['Variable definition', 'Parameter definition', 'Display value']
        found_it = False
        for i in range(len(variable_type_list)):
            var_type_from_list = variable_type_list[i]
            if variable_type.lower() == var_type_from_list.lower():
                variable_type = variable_type_list[i]
                found_it = True
                break
            #end if
        #end for
        if not found_it:
            raise RuntimeError('variable_type ' + variable_type + 'not recognized')
        #end if
        variable_sign_dict = {'Variable definition':' = ', 'Parameter definition':' << ', 'Display value':':'}
        variable_sign = variable_sign_dict[variable_type]
        #
        if variable_type == 'Display value':
            equation_str = equation_name + variable_sign
        else:
            equation_str = equation_name + variable_sign + equation_value
        #end if
        #
        expression_list = self.expression_list
        if equation_str in expression_list:
            YesNo = messagebox.askyesno('Add Equation','Equation ' + equation_str + ' exists. Remove existing equation ?')
            if YesNo:
                self.remove_equation(equation_str)
            #end if
        #end if
        #
        try:
            self._eqns_doc.Equations.Add2(equation_str, x_pos, y_pos)
        except:
            raise RuntimeError('Could not add equation')
        #end try
        #
#
#**********************************************************************************************
#
class _Equation():
    '''
    Operations on a single equation
    
    Parameters
    ----------
    expression: str
           Expression is the term used for the string associated with the equation. Must be in the form:
           "x = 2", "x << 2" or "x:"
            
    document: object
          schematic, system diagram or output equation documement object variable 
    '''
    def __init__(self, expression: str, document: object):
        self._initialize_equation(document, expression)
        #
    def _initialize_equation(self, document: object, expression: str):#------------------------------
        self._eqn_doc = document
        self._equation_name, self._equation_value = self._split_expression(expression)
        self._variable_type = self._determine_variable_type(expression)
        self._expression = expression
        self._eqn_index = self._determine_eqn_index()
        #
    @property
    def equation_name(self) -> str: #------------------------------------------
        '''returns equation name'''
        return self._equation_name
        #
    @property
    def equation_enabled(self) -> bool:#------------------------------------------
        return self._eqn_doc.Equations(self._eqn_index).Enabled
        #
    @equation_enabled.setter
    def equation_enabled(self, enabled_state: bool):#--------------------------
        if not isinstance(enabled_state, bool):
            raise TypeError('enabled_state must be boolean type')
        #end if
        self._eqn_doc.Equations(self._eqn_index).Enabled = enabled_state
        #
    @equation_name.setter
    def equation_name(self, equation_name: str):
        '''modify equation name'''
        if not isinstance(equation_name, str):
            raise TypeError('equation_name must be string type')
        #end if
        self._equation_name = equation_name
        self._update_equation()
        #
    @property
    def equation_value(self) -> str:#--------------------------------------------
        ''' returns equation value as string'''
        return self._equation_value
        #
    @equation_value.setter
    def equation_value(self, equation_value):#-----------------------------
        ''' set equation value'''
        if isinstance(equation_value, int) or isinstance(equation_value, float):
            equation_value = str(equation_value)
        elif not isinstance(equation_value, str):
            raise TypeError('equation_value type must be str, float or int')
        #end if
        self._equation_value = equation_value
        self._update_equation()
        #
    @property
    def variable_type(self) -> str:#------------------------------------------------
        '''returns variable type: Variable definition, Parameter definition, Display value'''
        return self._variable_type
        #
    @variable_type.setter
    def variable_type(self, variable_type: str):#------------------------------------
        '''sets the variable type'''
        variable_type_list = ['Variable definition', 'Parameter definition', 'Display value']
        found_it = False
        for i in range(len(variable_type_list)):
            var_type_from_list = variable_type_list[i]
            if variable_type.lower() == var_type_from_list.lower():
                self._variable_type = variable_type_list[i]
                self._update_equation()
                found_it = True
                break
            #end if
        #end for
        #
        if not found_it:
            raise RuntimeError('variable_type ' + variable_type + 'not recognized')
        #end if
        #
    @property
    def expression(self) -> str:#--------------------------------------------------
        ''' returns expression'''
        return self._expression
        #
    @property
    def xy_position(self) -> tuple:#-------------------------------------------------
        x_position = self._eqn_doc.Equations(self._eqn_index).x
        y_position = self._eqn_doc.Equations(self._eqn_index).y
        xy_position_tuple = (x_position, y_position)
        return xy_position_tuple
        #
    def set_xy_position(self, xy_position_tuple: tuple):#------------------------------
        '''set xy position'''
        self._eqn_doc.Equations(self._eqn_index).x = xy_position_tuple[0]
        self._eqn_doc.Equations(self._eqn_index).y = xy_position_tuple[1]
        #
    @property
    def optimize_enabled(self) -> bool:#----------------------------------------------
        '''returns Optimize enabled state '''
        return self._eqn_doc.Equations(self._eqn_index).Optimize
        #
    @optimize_enabled.setter
    def optimize_enabled(self, enabled: bool):#--------------------------------------
        '''sets Optimize enabled state '''
        self._eqn_doc.Equations(self._eqn_index).Optimize = enabled
        #
    @property
    def constrain(self) -> bool:#-------------------------------------------------------
        '''returns Contrain enabled state '''
        return self._eqn_doc.Equations(self._eqn_index).Constrain
        #
    @constrain.setter
    def constrain(self, enabled: bool):#----------------------------------------------------
        '''sets Contrain enabled state '''
        self._eqn_doc.Equations(self._eqn_index).Constrain = enabled
        #    
    @property
    def lower_constraint(self) -> float:#-------------------------------------------------
        ''' returns LowerContraint value'''
        return self._eqn_doc.Equations(self._eqn_index).LowerConstraint
        #
    @lower_constraint.setter
    def lower_constraint(self, lower_contraint_value:float):#-----------------------------
        ''' sets LowerContraint value'''
        if not isinstance(lower_contraint_value, float) and  not isinstance(lower_contraint_value, int):
            raise TypeError('lower_contraint_value must be float or int')
        #end if
        self._eqn_doc.Equations(self._eqn_index).LowerConstraint = lower_contraint_value
        # 
    @property
    def upper_constraint(self) -> float:#-------------------------------------------------
        ''' returns UpperContraint value'''
        return self._eqn_doc.Equations(self._eqn_index).UpperConstraint
        #
    @upper_constraint.setter
    def upper_constraint(self, upper_contraint_value:float):#-----------------------------
        ''' sets UpperContraint value'''
        if not isinstance(upper_contraint_value, float) and  not isinstance(upper_contraint_value, int):
            raise TypeError('upper_contraint_value must be float or int')
        #end if
        self._eqn_doc.Equations(self._eqn_index).UpperConstraint = upper_contraint_value
        #
    @property
    def step_size(self) -> float:#-------------------------------------------------
        ''' returns StepSize value'''
        return self._eqn_doc.Equations(self._eqn_index).StepSize
        #
    @step_size.setter
    def step_size(self, step_size_value:float):#-----------------------------
        ''' sets StepSize value'''
        if not isinstance(step_size_value, float) and  not isinstance(step_size_value, int):
            raise TypeError('step_size_value must be float or int')
        #end if
        self._eqn_doc.Equations(self._eqn_index).StepSize = step_size_value
        #        
    def _update_equation(self):#---------------------------------------------------------
        '''write equation to AWR document'''
        variable_sign_dict = {'Variable definition':' = ', 'Parameter definition':' << ', 'Display value':':'}
        variable_sign = variable_sign_dict[self._variable_type]
        if self._variable_type == 'Display value':
            self._expression = self._equation_name + variable_sign
        else:
            self._expression = self._equation_name + variable_sign + self._equation_value
        #end if
        self._eqn_doc.Equations(self._eqn_index).Expression = self._expression
        #
    def _determine_eqn_index(self):#-----------------------------------------------------
        '''returns the equation index in the collection of equations in the document'''
        for eqn_idx in range(1, self._eqn_doc.Equations.Count+1):
            eqn_expression = self._eqn_doc.Equations(eqn_idx).Expression
            eqn_name, eqn_value = self._split_expression(eqn_expression)
            eqn_var_type = self._determine_variable_type(eqn_expression)
            if (self._equation_name == eqn_name) and (self._variable_type == eqn_var_type):
                break
            #end if
        #end for
        return eqn_idx
        #
    def _split_expression(self, expression: str) -> str:#-----------------------------------------
        '''
        splits full equation into equation name and its value as string
        
        Parameters
        ----------
        expression: string
                 full expression from document
        
        Returns
        -------
        equation_name: string
        equation_value: string
        '''
        variable_type_list = ['=', '<<', ':']
        for i in range(len(variable_type_list)):
            if variable_type_list[i] in expression:
                variable_type_str = variable_type_list[i]
                break
            #end if
        #end if
        
        equation_split = expression.split(variable_type_str)
        equation_name = equation_split[0]
        equation_name = equation_name.lstrip()
        equation_name = equation_name.rstrip()
        if variable_type_str == ':':
            equation_value = ''
        else:
            equation_value = equation_split[1]
            equation_value = equation_value.lstrip()
            equation_value = equation_value.rstrip()
        #end if
        return equation_name, equation_value
        #
    def _determine_variable_type(self, expression: str):#----------------------------------------
        ''' returns variable type from the expression'''
        variable_type_dict = {'=':'Variable definition', '<<':'Parameter definition', ':':'Display value'}
        for variable_type_key in list(variable_type_dict):
            if variable_type_key in expression:
                variable_type_str = variable_type_dict[variable_type_key]
                break
            #end if
        #end if
        return variable_type_str
        #  
#
#**********************************************************************************************
#    
class _ProcessDefinitions():
    '''
    Methods related to Layout Process 
    
    Parmeters
    ---------
    awrde: object variable
         The AWRDE object returned from awrde_utils.EstablishLink()
         
    '''
    def __init__(self, awrde):#-------------------------------------------------
        self.awrde = awrde
        #
    def _unit_type_str(self, unit_type_enum: int) -> str:#-------------------------------------
        '''
        Returns a string that is used as a key for the dictionary returned from _project_units
        
        Parameters
        ----------
        unit_type_enum: integer
                   This value is returned from element parameter UnitType and measurement UnitType
        
        Returns
        -------
        unit_type_str: string
               String that corresponds to the dictionary keys returned from _project_units
        '''
        if mwo.mwUnitType(unit_type_enum) == mwo.mwUnitType.mwUT_Frequency:
            unit_type_str = 'Frequency'
        elif mwo.mwUnitType(unit_type_enum) == mwo.mwUnitType.mwUT_Capacitance:
            unit_type_str = 'Capacitance'
        elif mwo.mwUnitType(unit_type_enum) == mwo.mwUnitType.mwUT_Inductance:
            unit_type_str = 'Inductance'
        elif mwo.mwUnitType(unit_type_enum) == mwo.mwUnitType.mwUT_Resistance:
            unit_type_str = 'Resistance'
        elif mwo.mwUnitType(unit_type_enum) == mwo.mwUnitType.mwUT_Conductance:
            unit_type_str = 'Conductance'
        elif mwo.mwUnitType(unit_type_enum) == mwo.mwUnitType.mwUT_Length:
            unit_type_str = 'Length'
        elif mwo.mwUnitType(unit_type_enum) == mwo.mwUnitType.mwUT_Temperature:
            unit_type_str = 'Temperature'
        elif mwo.mwUnitType(unit_type_enum) == mwo.mwUnitType.mwUT_Angle:
            unit_type_str = 'Angle'
        elif mwo.mwUnitType(unit_type_enum) == mwo.mwUnitType.mwUT_Time:
            unit_type_str = 'Time'
        elif mwo.mwUnitType(unit_type_enum) == mwo.mwUnitType.mwUT_Voltage:
            unit_type_str = 'Voltage'
        elif mwo.mwUnitType(unit_type_enum) == mwo.mwUnitType.mwUT_Current:
            unit_type_str = 'Current'
        elif mwo.mwUnitType(unit_type_enum) == mwo.mwUnitType.mwUT_PowerLog:
            unit_type_str = 'Power in dB'
        elif mwo.mwUnitType(unit_type_enum) == mwo.mwUnitType.mwUT_Power:
            unit_type_str = 'Power'
        elif mwo.mwUnitType(unit_type_enum) == mwo.mwUnitType.mwUT_None:
            unit_type_str = 'None'
        else:
            warnings.warn('unit_type_enum not recognized')
        #end if
        return unit_type_str
        #
    def _project_units(self, lpf_name: str) -> dict:#---------------------------------------------
        '''
        returns dictionary of unit strings according to the layout process file LPF_Name
        
        Parameters
        ----------
        lpf_name: string
                Layout Process File name
        
        Returns
        -------
        lpf_units_dict: dictionary
              keys are the unit type
              values are the unit string for each unit type
              
        '''
        for lpf in self.awrde.Project.ProcessDefinitions:
            if lpf.Name == lpf_name:
                break
            #end if
        #end for
        #
        lpf_units_dict = {}
        lpf_units_dict['Frequency'] = lpf.Units.Item(mwo.mwUnitType.mwUT_Frequency).UnitString
        lpf_units_dict['Capacitance'] = lpf.Units.Item(mwo.mwUnitType.mwUT_Capacitance).UnitString
        lpf_units_dict['Inductance'] = lpf.Units.Item(mwo.mwUnitType.mwUT_Inductance).UnitString
        lpf_units_dict['Resistance'] = lpf.Units.Item(mwo.mwUnitType.mwUT_Resistance).UnitString
        lpf_units_dict['Conductance'] = lpf.Units.Item(mwo.mwUnitType.mwUT_Conductance).UnitString
        lpf_units_dict['Length'] = lpf.Units.Item(mwo.mwUnitType.mwUT_Length).UnitString
        lpf_units_dict['Temperature'] = lpf.Units.Item(mwo.mwUnitType.mwUT_Temperature).UnitString
        lpf_units_dict['Angle'] = lpf.Units.Item(mwo.mwUnitType.mwUT_Angle).UnitString
        lpf_units_dict['Time'] = lpf.Units.Item(mwo.mwUnitType.mwUT_Time).UnitString
        lpf_units_dict['Voltage'] = lpf.Units.Item(mwo.mwUnitType.mwUT_Voltage).UnitString
        lpf_units_dict['Current'] = lpf.Units.Item(mwo.mwUnitType.mwUT_Current).UnitString
        lpf_units_dict['Power in dB'] = lpf.Units.Item(mwo.mwUnitType.mwUT_PowerLog).UnitString
        lpf_units_dict['Power'] = lpf.Units.Item(mwo.mwUnitType.mwUT_Power).UnitString
        lpf_units_dict['None'] = 'None'
        return lpf_units_dict
        #
    def _unit_scale_factor(self, unit_type_str: str, unit_string: str) -> dict:#----------------------------------------------------
        '''
        returns a dictionary for applying offset and scaling to project units
        
        Parameters
        ----------
        unit_type_str: string
              String corresponding to dictionary key returned from _PojectUnits
              
        unit_string: string
               UnitType parameter from element or measurement
        
        Returns
        -------
        unit_scale_factor_dict: dictionary
               keys: Offset1, Offset2, Multiplier
               y = (x + Offset2)*Multiplier +  Offset1
               
        
        '''
        offset1_val = 0
        offset2_val = 0
        multiplier_val = 1
        #
        if unit_type_str == 'Frequency':
            freq_dict = {'THz':1e-12, 'GHz':1e-9, 'MHz':1e-6, 'kHz':1e-3, 'Hz':1}
            multiplier_val = freq_dict[unit_string]
            #
        elif unit_type_str == 'Capacitance':
            capacitance_dict ={'fF':1e15, 'pF':1e12, 'nF':1e9, 'uF':1e6, 'mF':1e3, 'F':1}
            multiplier_val = capacitance_dict[unit_string]
            #
        elif unit_type_str == 'Inductance':
            inductance_dict ={'fH':1e15, 'pH':1e12, 'nH':19, 'uH':1e6,
                              'mH':1e3, 'H':1, 'kH':1e-3, 'MH':1e-6, 'GH':1e-9, 'TH':1e-12}
            multiplier_val = inductance_dict[unit_string]
            #
        elif unit_type_str == 'Resistance':
            resistance_dict = {'fOhm':1e15, 'pOhm':1e12, 'nOhm':1e9, 'uOhm':1e6, 'mOhm':1e3,
                               'Ohm':1, 'kOhm':1e-3, 'MOhm':1e-6, 'GOhm':1e-9, 'TOhm':1e-12}
            multiplier_val = resistance_dict[unit_string]
            #
        elif unit_type_str == 'Conductance':
            conductance_dict = {'fS':1e15, 'pS':1e12, 'nS':1e9, 'uS':1e6, 'mS':1e3, 'S':1}
            multiplier_val = Conductancce_dict[unit_string]
            #
        elif unit_type_str == 'Length':
            length_dict = {'fm':1e15, 'pm':1e12, 'nm':1e9, 'um':1e6, 'mmm':1e3, 'cm':1e2, 'm':1, 'km':1e-3,
                           'mil':(100*1000/2.54), 'inch':(100/2.54), 'foot':(100/(2.54*12)), 'mile':(100/(2.54*12*5280))}
            multiplier_val = length_dict[unit_string]
            #
        elif unit_type_str == 'Temperature':
            temperature_mult_dict ={'DegC':1, 'DegK':1, 'DegF':(9/5)}
            temperature_offset1_dict ={'DegC':273.15, 'DegK':0, 'DegF':32}
            temperature_offset2_dict ={'DegC':0, 'DegK':0, 'DegF':273.15}
            multiplier_val = temperature_mult_dict[unit_string]
            offset1_val = TemperatureOffset_dict[unit_string]
            #
        elif unit_type_str == 'Angle':
            angle_dict = {'Rad':1, 'Deg':(180/np.pi)}
            multiplier_val = angle_dict[unit_string]
            #
        elif unit_type_str == 'Time':
            time_dict = {'fs':1e15, 'ps':1e12, 'ns':1e9, 'us':1e6, 'ms':1e3, 's':1, 'ks':1e-3, 'Ms':1e-6, 'Gs':1e-9, 'Ts':1e-12}
            multiplier_val = time_dict[unit_string]
            #
        elif unit_type_str == 'Voltage':
            voltage_dict = {'fV':1e15, 'pV':1e12, 'nV':1e9, 'mV':1e3, 'V':1, 'kV':1e-3, 'MV':1e-6, 'GV':1e-9, 'TV':1e-12}
            multiplier_val = voltage_dict[unit_string]
            #
        elif unit_type_str == 'Current':
            current_dict = {'fA':1e15, 'pA':1e12, 'nA':1e9, 'uA':1e6, 'mA':1e3, 'A':1, 'kA':1e-3, 'MA':1e-6, 'GA':1e-9, 'TA':1e-12}
            multiplier_val = current_dict[unit_string]
            #
        elif unit_type_str == 'Power in dB':
            power_offset_dict = {'dBm':30, 'dBW':0}
            offset1_val = power_offset_dict[unit_string]
            #
        elif unit_type_str == 'Power':
            power_dict = {'fW':1e-15, 'pW':1e-12, 'nW':1e-9, 'uW':1e-6, 'mW':1e-3, 'W':1, 'kW':1e3, 'MW':1e6, 'GW':1e9, 'TW':1e12}
            multiplier_val = power_dict[unit_string]
            #
        elif unit_type_str == 'None':
            pass
        else:
            warnings.warn('unit_type_str not recognized: ' + unit_type_str)
        #end if
        #    
        unit_scale_factor_dict = {}
        unit_scale_factor_dict['Offset1'] = offset1_val
        unit_scale_factor_dict['Offset2'] = offset2_val
        unit_scale_factor_dict['Multiplier'] = multiplier_val
        return unit_scale_factor_dict
        #
        